library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.STD_LOGIC_ARITH.ALL;
use IEEE.STD_LOGIC_UNSIGNED.ALL;

entity EPNCTL is
    Port ( CLK : in std_logic;
	 		  RST : in std_logic;

			  L   : out std_logic_vector(1 downto 0); -- FPGA self reset

	 		  DATA_IN  : in  std_logic_vector(31 downto 0);
           RE       : in  std_logic;
			  RE_BSY   : out std_logic;

			  DATA_OUT : out std_logic_vector(31 downto 0);
			  WE       : out std_logic;
			  WE_RDY   : in  std_logic;

			  HRST : out   std_logic; 
           HD   : inout std_logic_vector(15 downto 0);
           HA   : out   std_logic_vector(2 downto 0);
           HCS  : out   std_logic;
           HWR  : out   std_logic_vector(1 downto 0);
           HRD  : out   std_logic_vector(1 downto 0) );
end EPNCTL;

architecture RTL of EPNCTL is

signal STATE     : std_logic_vector( 4 downto 0 );
constant S0_IDLE : std_logic_vector( 4 downto 0 ):= "00000";
constant S0_RD   : std_logic_vector( 4 downto 0 ):= "00001";
constant S0_WR   : std_logic_vector( 4 downto 0 ):= "00010";
constant S1_RD   : std_logic_vector( 4 downto 0 ):= "00011";
constant S1_WR   : std_logic_vector( 4 downto 0 ):= "00100";
constant S2_RD   : std_logic_vector( 4 downto 0 ):= "00101";
constant S2_WR   : std_logic_vector( 4 downto 0 ):= "00110";
constant S3_RD   : std_logic_vector( 4 downto 0 ):= "00111";
constant S3_WR   : std_logic_vector( 4 downto 0 ):= "01000";
constant S4_RD   : std_logic_vector( 4 downto 0 ):= "01001";
constant S4_WR   : std_logic_vector( 4 downto 0 ):= "01010";
constant S5_RD   : std_logic_vector( 4 downto 0 ):= "01011";
constant S5_WR   : std_logic_vector( 4 downto 0 ):= "01100";

constant S0_EPNRST  : std_logic_vector( 4 downto 0 ):= "01101";
constant S1_EPNRST  : std_logic_vector( 4 downto 0 ):= "01110";
constant S2_EPNRST  : std_logic_vector( 4 downto 0 ):= "01111";

constant S0_FPGARST : std_logic_vector( 4 downto 0 ):= "10000";
constant S1_FPGARST : std_logic_vector( 4 downto 0 ):= "10001";
constant S2_FPGARST : std_logic_vector( 4 downto 0 ):= "10010";

constant NUL    : std_logic_vector( 7 downto 0 ) := (X"00");
constant SYN    : std_logic_vector( 7 downto 0 ) := (X"16");
constant RDFLG  : std_logic_vector( 3 downto 0 ) := (X"A");
constant WRFLG  : std_logic_vector( 3 downto 0 ) := (X"5");

constant FPGA_RST : std_logic_vector( 7 downto 0 ) := (X"1A");
constant EPN_RST  : std_logic_vector( 7 downto 0 ) := (X"1B");


signal HD_REG  : std_logic_vector(15 downto 0 );
signal DIR     : std_logic;

signal RST_CNT : std_logic_vector(15 downto 0);

begin


HD <= HD_REG(15 downto 0) when( DIR = '1') else "ZZZZZZZZZZZZZZZZ";

--***                 *--
--**  STATE MACHINE  **--
--*                 ***--
process( CLK, RST ) begin
	if ( RST ='1') then
		L        <= "00"; -- FPGA self-reset
		WE       <= '0';
		DIR      <= '0';
		HCS      <= '1';
		HWR      <= "11";
		HRD      <= "11";
		RE_BSY   <= '0';
		HRST     <= '1'; 
		HA       <= (others=>'0');
		HD_REG   <= (others=>'0');
		DATA_OUT <= (others=>'0');
		RST_CNT  <= (others=>'0');
		STATE    <= S0_IDLE;
	elsif( CLK'event and CLK='1') then
		case STATE is
			when S0_IDLE =>
				-- FPGA SELF-RESET
				if( RE='1'                         and
					 DATA_IN(23 downto 16)=FPGA_RST and
					 DATA_IN(31 downto 24)=SYN ) then

					L <= "00";
               STATE <= S0_FPGARST;
			   -- EPN RESET
				elsif( RE='1'                     and
					 DATA_IN(23 downto 16)=EPN_RST and
					 DATA_IN(31 downto 24)=SYN ) then

					HRST  <= '0'; 
               STATE <= S0_EPNRST;
				-- READ         
				elsif( RE='1'                      and
					 DATA_IN(23 downto 20)=RDFLG and
					 DATA_IN(19)='0'             and
					 DATA_IN(31 downto 24)=SYN ) then

					RE_BSY <= '0';
					HA(2 downto 0) <= DATA_IN(18 downto 16);
					STATE  <= S0_RD;
				-- WRITE
				elsif( RE='1'                      and
				       DATA_IN(23 downto 20)=WRFLG and
						 DATA_IN(19)='0'             and
					    DATA_IN(31 downto 24)=SYN ) then

					RE_BSY <= '0';
					HD_REG <= DATA_IN(15 downto 0);
					HA(2 downto 0) <= DATA_IN(18 downto 16);
					STATE <= S0_WR;
				else
					STATE <= S0_IDLE;
				end if;
			--***                      *--
			--**  Read from Register  **--
			--*                      ***--
			when S0_RD =>
				HCS   <= '0';
				HRD   <= "00";
				HWR   <= "11";	
				STATE <= S1_RD;
			when S1_RD =>
				DATA_OUT(15 downto 0) <= HD;
				STATE <= S2_RD;
			when S2_RD =>
				DATA_OUT(31 downto 24) <= SYN;
				DATA_OUT(23 downto 16) <= NUL;
				HCS   <= '1';
				HRD   <= "11";
				HWR   <= "11";
				WE    <= '1';
				STATE <= S3_RD;
			when S3_RD =>
				WE      <= '0';
				RE_BSY  <= '0';

				if( WE_RDY='1') then
					STATE <= S0_IDLE;
				else
					STATE <= S3_RD;
				end if;
			--***                     *--
			--**  Write to Register  **--
			--*                     ***--
			when S0_WR =>
				HCS   <= '0';
				HRD   <= "11";
				HWR   <= "00";
				STATE <= S1_WR;
			when S1_WR =>
				DIR   <= '1';  -- output
				STATE <= S2_WR;
			when S2_WR =>
				HCS   <= '1';
				HRD   <= "11";
				HWR   <= "11";
				DIR   <= '0';  -- Hi-Z
				STATE <= S3_WR;
			when S3_WR =>
				STATE <= S0_IDLE;
			--***             *--
			--**  FPGA RESET  **--
			--*             ***--
			when S0_FPGARST => L <= "01"; STATE <= S1_FPGARST;
			when S1_FPGARST => L <= "10"; STATE <= S2_FPGARST;
			when S2_FPGARST => L <= "11"; STATE <= S0_IDLE;
			--***             *--
			--**  EPN RESET  **--
			--*             ***--
			when S0_EPNRST =>
				if( RST_CNT=(X"FFFF")) then
					HRST    <= '1';
					RST_CNT <= (others=>'0');
					STATE   <= S0_IDLE;
				else
					RST_CNT <= RST_CNT + '1';
					STATE   <= S0_EPNRST;
				end if;

			when others => STATE <= S0_IDLE;
		end case;
	end if;
end process;

end RTL;

